####################################################################
# Copyright (c) Kapil Thangavelu. All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
####################################################################

from sqlalchemy import orm
from zope import interface

from ore.contentmirror import schema, interfaces

class PeerFactory( object ):
    
    interface.implements( interfaces.IPeerFactory )
    
    def __init__( self, context, transformer ):
        self.context = context
        self.transformer = transformer
        
    @property
    def name(self):
        return self.context.__class__.__name__ + 'Peer'
        
    def make( self ):
        klass = type( self.name, (schema.Content,),
                      dict(transformer=self.transformer) )

        orm.mapper( klass, 
                    self.transformer.table,
                    properties=dict(self.transformer.properties),
                    inherits=schema.Content,
                    polymorphic_on=schema.content.c.type,
                    polymorphic_identity=self.name )
        return klass
        
class PeerRegistry( object ):
    
    interface.implements( interfaces.IPeerRegistry )
    
    def __init__( self ):
        self._peer_classes = {}
        
    def __setitem__( self, key, value ):
        self._peer_classes[ key ] = value        
        
    def __getitem__( self, key ):
        return self._peer_classes[ key ]
        
    def __contains__( self, key ):
        return key in self._peer_classes

    def items( self ):
        return self._peer_classes.items()



####################################################################
# Copyright (c) Kapil Thangavelu. All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
####################################################################

"""
$Id: $
"""

import unittest
from zope.testing import doctest

def test_suite():
    import testing
    return unittest.TestSuite((
        doctest.DocFileSuite(
            'readme.txt',
            setUp=testing.setUp, tearDown=testing.tearDown,
            optionflags=doctest.NORMALIZE_WHITESPACE|doctest.ELLIPSIS,
            globs=testing.doctest_ns
            ),
        doctest.DocFileSuite(
            'ref.txt',
            setUp=testing.setUp, tearDown=testing.tearDown,
            optionflags=doctest.NORMALIZE_WHITESPACE|doctest.ELLIPSIS,
            globs=testing.doctest_ns
            ),           

    ))




####################################################################
# Copyright (c) Kapil Thangavelu. All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
####################################################################
"""

cms content model

"""
import os
from zope import interface
from sqlalchemy import orm, create_engine
import sqlalchemy as rdb

from ore.contentmirror.session import Session
from ore.contentmirror import interfaces

metadata =  rdb.MetaData()

if os.environ.get('CONTENTMIRROR_URI'):
    metadata.bind=create_engine(os.environ['CONTENTMIRROR_URI'])
    
ContentSequence = rdb.Sequence('content_sequence')

content = rdb.Table(
   "content",
   metadata,
   rdb.Column( "content_id", rdb.Integer, ContentSequence, primary_key=True ),
   rdb.Column( "id", rdb.String(256), nullable=False ),
   rdb.Column( "uid", rdb.String(36), nullable=False ),
   rdb.Column( "portal_type", rdb.String(64) ),
   rdb.Column( "status", rdb.String(64) ),   
   rdb.Column( "type", rdb.String(64) ),
   rdb.Column( "container_id", rdb.Integer, rdb.ForeignKey('content.content_id', ondelete="CASCADE")),
   rdb.Column( "path", rdb.Text ), 
   # dublin core
   rdb.Column( "title", rdb.Text ),
   rdb.Column( "description", rdb.Text ),
   rdb.Column( "subject", rdb.Text ),
   rdb.Column( "location", rdb.Text(4000) ),
   rdb.Column( "contributors", rdb.Text ),
   rdb.Column( "creators", rdb.Text ),
   rdb.Column( "creation_date", rdb.DateTime ),
   rdb.Column( "modification_date", rdb.DateTime ),
   rdb.Column( "effectivedate", rdb.DateTime  ),
   rdb.Column( "expirationdate", rdb.DateTime ),
   rdb.Column( "language", rdb.Text(32)  ),
   rdb.Column( "rights", rdb.Text ),
   # plone
   rdb.Column( "excludefromnav", rdb.Boolean)
   )

rdb.Index('content_uid_idx', content.c.uid, unique=True )
   
class Content( object ):
    type = "content"
    interface.implements( interfaces.IContentPeer )

orm.mapper( Content, content, 
            polymorphic_on=content.c.type,
            polymorphic_identity='content',
            properties = { 'children':orm.relation(Content,
                    backref=orm.backref('parent', remote_side=[content.c.content_id])) }
    )           

#class Node( object ): pass   
#orm.mapper( Node, content )
            
class UIDFilter( object ):

    __slots__ = ('uid',)
    
    def __init__( self, uid ):
        self.uid = uid
        
    def __call__( self, ob ):
        if not interfaces.IContentPeer.providedBy( ob ):
            return False
        return ob.uid == self.uid

def fromUID( content_uid ):
    """
    fetch an object by uid, does not flush the session, and also checks 
    against new objects in the sqlalchemy session.
    """
    session = Session()
    peers = filter( UIDFilter( content_uid ), session.new)
    if peers:
        return peers.pop()
    return session.query( Content ).autoflush(False).filter( content.c.uid == content_uid ).first()
                
relations = rdb.Table(
   "relations",
   metadata,
   rdb.Column( "source_id", rdb.Integer, 
               rdb.ForeignKey('content.content_id', ondelete='CASCADE'),
               primary_key=True ),
   rdb.Column( "target_id", rdb.Integer, 
               rdb.ForeignKey('content.content_id', ondelete='CASCADE'),
               primary_key=True),
   rdb.Column( "relationship", rdb.String(128), primary_key=True )
   )


class Relation( object ):

    def __init__( self, source=None, target=None, relation=None):
        self.source = source
        self.target = target
        self.relationship = relation
   
orm.mapper( Relation, relations,
            properties = {
                'source': orm.relation(Content, uselist=False, backref='relations',
                                       primaryjoin=content.c.content_id==relations.c.source_id ),
                'target': orm.relation(Content, uselist=False,
                                       primaryjoin=content.c.content_id==relations.c.target_id ),
                }
            )

files = rdb.Table(
   "files",
   metadata,
   rdb.Column( "content_id", rdb.Integer, 
        rdb.ForeignKey('content.content_id', ondelete="CASCADE"), primary_key=True ),
   rdb.Column( "attribute", rdb.String(156), primary_key=True ),           
   rdb.Column( "type", rdb.String(30) ),        
   rdb.Column( "content", rdb.Binary),           
   rdb.Column( "path", rdb.String(300) ),
   rdb.Column( "size", rdb.Integer ),      
   rdb.Column( "checksum", rdb.String(33) ),   
   rdb.Column( "file_name", rdb.String(156) ),
   rdb.Column( "mime_type", rdb.String(80) ),
   )
   
rdb.Index('files_idx', 
      files.c.content_id,
      files.c.attribute,
      unique=True )
      
class File( object ): 
    """ """
    
orm.mapper( File, files,
        polymorphic_on=files.c.type,
        polymorphic_identity='db-file' )




####################################################################
# Copyright (c) Kapil Thangavelu. All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
####################################################################
"""
The Model Loader drives the startup bootstrapping, its wired into the
zcml directives, its responsible for loading/generating the tables, 
and creating the rdb mapped peer classes.

"""
from zope import component
from ore.contentmirror import schema, interfaces

class ModelLoader( object ):
    
    def __init__( self, metadata ):
        self.metadata = metadata

    def load( self, klass ):
        instance = klass("transient")

        registry = component.queryUtility( interfaces.IPeerRegistry )
        if klass in registry:
            raise KeyError("Duplicate %r"%klass)

        transformer = self.transform( instance )
        peer_class = self.peer( instance, transformer )
        
        registry[ klass ] = peer_class
        
    def transform( self, instance ):
        transformer = component.getMultiAdapter( (instance, self.metadata ),
                                                 interfaces.ISchemaTransformer )
        transformer.transform()
        return transformer
        
    def peer( self, instance, transformer ):
        factory = component.getMultiAdapter( ( instance, transformer ) ,
                                             interfaces.IPeerFactory )
        return factory.make()
        
loader = ModelLoader( schema.metadata )
load = loader.load

   



####################################################################
# Copyright (c) Kapil Thangavelu. All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
####################################################################
"""
Testing Infrastructure
"""

from zope import interface, component
from zope.component import testing as placelesssetup

import time, random, md5, sqlalchemy
import interfaces, transform, peer, schema, serializer, operation

interface.classImplements( sqlalchemy.MetaData, interfaces.IMetaData )
                           
def setUp( test ):
    placelesssetup.setUp()
    component.provideAdapter( transform.StringTransform )
    component.provideAdapter( transform.IntegerTransform )
    component.provideAdapter( transform.FloatTransform )
    component.provideAdapter( transform.DateTimeTransform )
    component.provideAdapter( transform.LinesTransform )    
    component.provideAdapter( transform.BooleanTransform )
    component.provideAdapter( transform.FileTransform )
    component.provideAdapter( transform.PhotoTransform )
    component.provideAdapter( transform.ReferenceTransform )


    component.provideUtility( peer.PeerRegistry() )
    component.provideAdapter( peer.PeerFactory,
                              ( interfaces.IMirrored, interfaces.ISchemaTransformer ) )
    component.provideAdapter( transform.SchemaTransformer, 
                              ( interfaces.IMirrored, interfaces.IMetaData ) )
    component.provideAdapter( serializer.Serializer, (interfaces.IMirrored,) )
    component.provideAdapter( operation.OperationFactory, ( interfaces.IMirrored,) )
    component.provideUtility( operation.OperationBufferFactory() )
                              
    
def tearDown( test ):
    placelesssetup.tearDown()

def make_uuid(*args):
    t = str(time.time() * 1000L)
    r = str(random.random()*100000000000000000L)
    data = t +' '+ r +' '+ str(random.random()*100000000000000000L)+' '+ str(args)
    uid = md5.md5(data).hexdigest()
    return uid

class WorkflowTool( object ):

    def getCatalogVariablesFor( self, instance ):
        return {'review_state':getattr( instance, 'workflow_state', 'published') }
        
class BaseContent( object ):

    interface.implements( interfaces.IPortalContent )
    
    portal_type = ""
    schema = None
    portal_workflow = WorkflowTool()
    
    def __init__( self, id, container=None, **kw ):
        self.id = id
        self.uid = make_uuid(id)
        
        for k,v in kw.items():
            if k in self.schema:
                setattr(self,k,v)
                
        self.container = container
        
    def Schema( self ):
        return self.schema
        
    def UID( self ):
        return self.uid
        
    def getParentNode( self ):
        return self.container

    def getPhysicalPath( self ):
        path = []
        ob = self
        while True:
            path.append( ob.id )
            ob = ob.container
            if ob is None:
                break
        path.reverse()
        return path
    
class Schema( object ):

    def __init__(self, fields ):
        self._fields = fields
        
    def fields( self ):
        return self._fields
        
    def __contains__( self, name ):
        for f in self._fields:
            if f.__name__ == name:
                return True

class DateTime( object ):
    # mock zope2 datetime
    def __init__( self ):
        self.value = time.time()
    
    def timeTime( self ):
        return self.value
        

class File( object ):
    
    def __init__( self, id, content, mime_type="text/plain"):
        self.id = id
        self.data = content
        self.mime_type = mime_type
    
    def getId( self ):
        return self.id     
        
    
    
class MockField( object ):
    
    defaults = { 'required':False, 'default':'' }
        
    def __init__( self, name=None, **kw ):

        self.__name__ = name
        # copy
        values = dict(self.defaults)
        values.update(kw)
        self.__dict__.update( values )
        
    def getAccessor( self, instance ):
        return lambda : getattr( instance, 
                                 self.__name__, 
                                 self.default )
    
class StringField( MockField ):
    interface.implements( interfaces.IStringField )  

class IntegerField( MockField ):
    interface.implements( interfaces.IIntegerField )  

class FloatField( MockField ):
    interface.implements( interfaces.IFloatField )  
    
class ReferenceField( MockField ):
    multiValued = True
    interface.implements( interfaces.IReferenceField )  
    
class LinesField( MockField ):
    interface.implements( interfaces.ILinesField )  
    
class TextField( MockField ):
    interface.implements( interfaces.ITextField )  

class FileField( MockField ):
    interface.implements( interfaces.IFileField )  

    def getContentType( self, instance ):
        return self.getAccessor( instance )().mime_type

class BooleanField( MockField ):
    interface.implements( interfaces.IBooleanField )  

class ImageField( MockField ):
    interface.implements( interfaces.IImageField )  
    
class DateTimeField( MockField ):
    interface.implements( interfaces.IDateTimeField )  

class PhotoField( MockField ):
    interface.implements( interfaces.IPhotoField )  
    
doctest_ns = {
    'IntegerField'  : IntegerField,
    'FloatField'    : FloatField,
    'TextField'     : TextField,
    'StringField'   : StringField,
    'FileField'     : FileField,    
    'ImageField'    : ImageField,
    'DateTimeField' : DateTimeField,
    'ReferenceField': ReferenceField,
    'LinesField'    : LinesField,
    'BaseContent'   : BaseContent,
    'Schema'        : Schema,
    'DateTime'      : DateTime,
    'File'          : File
    }        



####################################################################
# Copyright (c) Kapil Thangavelu. All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
####################################################################

from zope import interface, component
from ore.contentmirror import schema, interfaces
from ore.contentmirror.session import Session

class Serializer( object ):

    interface.implements( interfaces.ISerializer )
    
    def __init__( self, context ):
        self.context = context
                
    def add( self ):
        registry = component.getUtility( interfaces.IPeerRegistry )
        peer = registry[ self.context.__class__ ]()
        session = Session()
        session.add( peer )        
        self._copy( peer )
        return peer
        
    def update( self ):
        peer = schema.fromUID( self.context.UID() )
        if peer is None:
            return self.add()
        self._copy( peer )
        return peer
        
    def delete( self ):
        peer = schema.fromUID( self.context.UID() )
        if peer is None:
            return
        session = Session()
        session.delete( peer )
        session.flush()

    def _copy( self, peer ):
        self._copyPortalAttributes( peer )
        peer.transformer.copy( self.context, peer )        
        self._copyContainment( peer )
        
    def _copyPortalAttributes( self, peer ):
        peer.portal_type = self.context.portal_type
        peer.uid = self.context.UID()
        peer.id  = self.context.id

        peer.path = '/'.join( self.context.getPhysicalPath() )

        wf_tool = getattr( self.context, 'portal_workflow', None )
        if wf_tool is None: return
        peer.status = wf_tool.getCatalogVariablesFor( self.context ).get('review_state')
        
    def _copyContainment( self, peer ):
        container = self.context.getParentNode()
        if container is None:
            return
        uid = getattr( container, 'UID', None)
        if uid is None: return
        uid = uid()
        container_peer = schema.fromUID( uid )
        if not container_peer:
            serializer = interfaces.ISerializer( container, None )
            if not serializer: return
            container_peer = serializer.add()
        peer.parent = container_peer

        
    
        



####################################################################
# Copyright (c) Kapil Thangavelu. All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
####################################################################

import threading, transaction
from zope import interface, component
from ore.contentmirror.session import Session
from ore.contentmirror import interfaces

class Operation( object ):
    
    interface.implements( interfaces.IOperation )

    filtered = False
    
    def __init__( self, context ):
        self.context = context
        
    @property
    def document_id( self ):
        return self.context.UID()
        
    def process( self ):
        raise NotImplementedError("subclass responsibility")

class AddOperation( Operation ):

    interface.implements( interfaces.IAddOperation )

    def process( self ):
        interfaces.ISerializer( self.context).add()
        
class UpdateOperation( Operation ):

    interface.implements( interfaces.IUpdateOperation )

    def process( self ):
        interfaces.ISerializer( self.context).update()
        
class DeleteOperation( Operation ):
    
    interface.implements( interfaces.IDeleteOperation )
    
    def process( self ):
        interfaces.ISerializer( self.context ).delete()
        
class OperationFactory( object ):

    interface.implements( interfaces.IOperationFactory )

    __slots__ = ('context',)

    def __init__( self, context ):
        self.context = context
    
    def add( self ):
        return self._store( AddOperation( self.context ) )

    def update( self ):
        return self._store( UpdateOperation( self.context ) )

    def delete( self ):
        return self._store( DeleteOperation( self.context ) )
        
    def _store( self, op ):
        # apply filters to operation
        component.subscribers( (self.context, op ), interfaces.IFilter )
        if op.filtered:
            return
        get_buffer().add( op )
        
class BufferManager( object ):

    def __init__( self, context ):
        self.context = context

    def abort( self, *args ):
        self.context.clear()

    def nothing( self, *args ): pass

    def sortKey( self ):
        return "content-mirror-2008"
    
    tpc_abort = abort
    commit = tpc_begin = tpc_vote = tpc_finish = nothing

class OperationBufferFactory( object ):
    # we componentize this so we can plugin alternate implementations
    # that can do async processing
    interface.implements( interfaces.IOperationBufferFactory )
    
    def new( self ):
        return OperationBuffer()
    
class OperationBuffer( object ):
    """
    an operation buffer aggregates operations across a transaction
    """

    interface.implements( interfaces.IOperationBuffer )
    
    def __init__( self ):
        self.ops = {}
        self.registered = False

    def add( self, op ):
        """add an operation to the buffer, aggregating with existing operations"""
        doc_id = op.document_id
        previous = self.ops.get( doc_id )
        if previous is not None:
            op = self._choose( previous, op)
        if op is None: # none returned when ops cancel
            del self.ops[ doc_id ]
            return
        self.ops[ doc_id ] = op
        if not self.registered:
            self._register()

    def clear( self ):
        self.ops = {}
        self.registered = False
        self.manager = None

    def flush( self ):
        for op in self:
            op.process()
        Session().flush()
        self.clear()
        
    def get( self, doc_id ):
        return self.ops.get( doc_id )
        
    def __iter__( self ):
        for op in self.ops.values():
            yield op
            
    def _register( self ):
        self.registered = True
        transaction.get().addBeforeCommitHook( self.flush )
        transaction.get().join( BufferManager( self ) )

    def _choose( self, previous, new ):
        """
        for a given content object, choose one operation to perform given
        two candidates. can also return no operations.
        """
        p_kind = (interfaces.IDeleteOperation.providedBy( previous ) and 2) \
                 or (interfaces.IAddOperation.providedBy( previous ) and 1) \
                 or (interfaces.IUpdateOperation.providedBy( previous ) and 0)
                 
        n_kind = (interfaces.IDeleteOperation.providedBy( new ) and 2) \
                 or (interfaces.IAddOperation.providedBy( new ) and 1) \
                 or (interfaces.IUpdateOperation.providedBy( new ) and 0)

        # if we have an add and then a delete, its an effective no-op
        if ( p_kind == 1 and n_kind == 2 ):
            return None
        if p_kind > n_kind:
            return previous
        return new

_buffer = threading.local()

def get_buffer( ):
    op_buffer = getattr( _buffer, 'buffer', None)
    if op_buffer is not None:
        return op_buffer
    op_buffer = component.getUtility( interfaces.IOperationBufferFactory ).new()
    _buffer.buffer = op_buffer
    return op_buffer





config.update( dict(
    project = "ore.contentmirror",
    version = open('version.txt').read().strip(),
    repo_url = "https://contentmirror.googlecode.com/svn"
    ))

def release( ):
    egg_tag()
    egg_release()
    product_tag()
    product_build( config.version )
    product_release()

def egg_tag( ):
    local('svn cp %(repo_url)s/%(project)s/trunk %(repo_url)s/%(project)s/tags/%(version)s -m "tagging version %(version)s"')
    
def egg_release( ):
    local('python setup.py register')
    local('python setup.py sdist upload')

def product_build( version='trunk' ):
    import fnmatch, os, shutil  
    
    # remove previous build
    version = version.strip()   
        
    if os.path.exists('build/ContentMirror'):
        shutil.rmtree('build/ContentMirror')
        
    if version == 'trunk':
        local('svn export %(repo_url)s/Products.ContentMirror/trunk/Products/ContentMirror build/ContentMirror')  
    else:
        local('svn export %(repo_url)s/Products.ContentMirror/tags/%(version)s/Products/ContentMirror build/ContentMirror')
    
    # copy sqlalchemy egg
    sa_eggs = fnmatch.filter( os.listdir('eggs'), 'SQLAlchemy*' )
    sa_eggs.sort()
    shutil.copytree( 'eggs/%s'%(sa_eggs[-1]), 'build/ContentMirror/eggs/%s'%(sa_eggs[-1]))
    
    # copy zope.sqlalchemy egg
    zsa_eggs = fnmatch.filter( os.listdir('eggs'), 'zope.sqlalchemy*' )
    zsa_eggs.sort()
    shutil.copytree( 'eggs/%s'%(zsa_eggs[-1]), 'build/ContentMirror/eggs/%s'%(zsa_eggs[-1]))
        
    # unpack bdist egg 
    config.cm_egg = 'ore.contentmirror-%(version)s-py2.4.egg'
    
    if os.path.exists('build/%(cm_egg)s'%config):
        shutil.rmtree('build/%(cm_egg)s'%config)

    config.working_dir = os.path.abspath('.')
    os.mkdir('build/%(cm_egg)s'%(config))

    local('MACOSX_DEPLOYMENT_TARGET="10.3" python2.4 setup.py bdist_egg')
    local('cd build/%(cm_egg)s && unzip %(working_dir)s/dist/%(cm_egg)s')
    local('cp -Rf build/%(cm_egg)s build/ContentMirror/eggs')
    #shutil.copyfile('version.txt', 'build/ContentMirror/version.txt')
    
    local('cd build && tar czvf ContentMirror-%(version)s.tar.gz ContentMirror')

def product_tag( ):
    config.project = "Products.ContentMirror"
    local('svn cp %(repo_url)s/ore.contentmirror/trunk/version.txt %(repo_url)s/%(project)s/trunk/ -m "updating version to %(version)s"')            
    local('svn cp %(repo_url)s/%(project)s/trunk %(repo_url)s/%(project)s/tags/%(version)s -m "tagging version %(version)s"')
    
def product_release( ):
    local('googlecode_upload.py -s "Product Release %(version)s" -p contentmirror build/ContentMirror-%(version)s.tar.gz')
    



####################################################################
# Copyright (c) Kapil Thangavelu. All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
####################################################################
"""
Create the database structure for a content mirror and output to standard out.
"""

from StringIO import StringIO
from ore.contentmirror import schema
import sqlalchemy as rdb
import sys, optparse

HELP = "zopectl|instance run ddl.py database_type" 

def main( ):
    
    parser = optparse.OptionParser()
    parser.add_option('-d', '--drop', dest="drop", action="store_true", default=False,
                      help="Generate Drop Tables DDL")
    parser.add_option('-n', '--nocreate', dest="create", action="store_true", default=False,
                      help="Do Not Generate Create Tables DDL")
                      
    (options, args) = parser.parse_args()
                                 
    if not len( sys.argv ) == 2:
        print HELP
        sys.exit(1)

    db_type = sys.argv[1].strip()
    buf = StringIO()    

    def write_statement( statement, parameters=''):
        ddl = statement + parameters
        buf.write( ddl.strip() + ';\n\n' )
        
    db = rdb.create_engine('%s://'%(db_type),
                           strategy='mock',
                           executor=write_statement )
    if options.drop:
        schema.metadata.drop_all( db )
    if not options.create:
        schema.metadata.create_all( db )
    print buf.getvalue()
    
if __name__ == '__main__':
    main()



####################################################################
# Copyright (c) Kapil Thangavelu. All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
####################################################################

from sqlalchemy import orm
import sqlalchemy as rdb
from zope import interface, component

from md5 import md5
from datetime import datetime
from ore.contentmirror.session import Session
from ore.contentmirror import interfaces, schema

class SchemaTransformer( object ):
    """
    the schema transformer is stateful and in memory persistent, its responsible for
    parsing the at schema and transforming it to a database table.  

    its also responsible for copying an object state from an instance to a peer.
    """
    interface.implements( interfaces.ISchemaTransformer )
    
    def __init__( self, context, metadata):
        self.context = context
        self.metadata = metadata
        self.table = None
        self.properties = {}
        
    def transform( self ):
        columns = list(self.columns())
        self.table = rdb.Table( self.name,
                                self.metadata,
                                useexisting=True,
                                *columns )
        return self.table
                                
    def copy( self, instance, peer ):
        for field in self.context.Schema().fields():
            transformer = component.queryMultiAdapter( ( field, self ))
            transformer.copy( instance, peer )

    @property
    def name( self ):
        return self.context.__class__.__name__.lower().replace(' ', '_')        
        
    def filter( self, field ):
        # return false if filter, true if ok, filter dublin core
        return not field.__name__ in interfaces.DUBLIN_CORE

    def columns( self):
        yield rdb.Column( "content_id", 
                          rdb.Integer, 
                          rdb.ForeignKey('content.content_id', ondelete="CASCADE"), 
                          primary_key=True )

        for field in self.context.Schema().fields():
            if not self.filter( field ):
                continue
                
            transformer = component.getMultiAdapter( ( field, self ))
            result = transformer.transform()
            
            if isinstance( result, rdb.Column ):
                yield result
                

class BaseFieldTransformer( object ):
    
    interface.implements( interfaces.IFieldTransformer )

    column_type = rdb.String
    column_args = ()
    
    def __init__( self, context, transformer):
        self.context = context
        self.transformer = transformer

    def transform( self ):
        args, kw = self._extractDefaults()
        column = rdb.Column( self.name, 
                             self.column_type( *self.column_args ),  
                             *args,
                             **kw )
        self.transformer.properties[ self.name ] = column
        return column
    
    def copy( self, instance, peer ):
        accessor = self.context.getAccessor( instance )
        value = accessor()
        setattr( peer, self.name, value )
                                 
    @property                      
    def name( self ):
        name = self.context.__name__.lower()
        name = name.replace(' ', '_')
        if name in self._reserved_names:
            name = "at_" + name
        return name
    
    @property
    def _reserved_names(self):
        e = self.transformer.metadata.bind
        if e is None: return ()
        return e.dialect.preparer.reserved_words
                
    def _extractDefaults( self ):
        args = []
        #if use_field_default and field.default:
        #    args.append( rdb.PassiveDefault( field.default ) )
        kwargs = {} #{
        #'nullable' : not self.context.required,
        #    'key' : self.context.getName(),            
        #    }
        return args, kwargs        

class ComputedTransform( BaseFieldTransformer):
    component.adapts( interfaces.IComputedField, interfaces.ISchemaTransformer)
    column_type = rdb.Text
    column_args = ()
    
class StringTransform( BaseFieldTransformer ):    
    component.adapts( interfaces.IStringField, interfaces.ISchemaTransformer )
    column_type = rdb.Text
    column_args = ()
    
    def copy( self, instance, peer ):
        accessor = self.context.getAccessor( instance )
        value = accessor()
        
        # one of a number of common bad practices that plone permits is specing a widget
        # which doesn't correspond to the field. ie. using a lines widget with a string field
        # actually generates a list value for the field. sadly this sort of sloppy usage appears
        # to be rather common place, so we'll try to handle it inline.. at least for now. 
        if isinstance( value, (tuple, list)):
            return LinesTransform( self.context, self.transformer ).copy( instance, peer )
        # at least morph it into a value which won't cause further issues.
        if not isinstance( value, basestring):
            value = str( value )
            
        setattr( peer, self.name, value )    

class TextTransform( BaseFieldTransformer ):    
    component.adapts( interfaces.IStringField, interfaces.ISchemaTransformer )
    column_type = rdb.Text
    column_args = ()    
        
class LinesTransform( StringTransform ):    
    component.adapts( interfaces.ILinesField, interfaces.ISchemaTransformer )
    
    def copy( self, instance, peer ):
        accessor = self.context.getAccessor( instance )
        value = accessor()
        if isinstance(value, (list, tuple)):
            value = "\n".join(value)
        setattr( peer, self.name, value )        
        
class FileTransform( object ):    
    """
    a file field serializer that utilizes a file peer.
    """
    interface.implements( interfaces.IFieldTransformer )
    component.adapts( interfaces.IFileField, interfaces.ISchemaTransformer )    

    def __init__( self, context, transformer ):
        self.context = context
        self.transformer = transformer
        
    def transform( self ):
        file_orm = orm.relation(schema.File,
                                cascade='all',
                                uselist=False,
                                primaryjoin=rdb.and_( 
                                    schema.files.c.content_id == schema.content.c.content_id,
                                    schema.files.c.attribute  == self.name )
                                )
        self.transformer.properties[ self.name ] = file_orm

    def copy( self, instance, peer ):
        accessor = self.context.getAccessor( instance )
        value = accessor()
        if not value: return 
        file_peer =self._getPeer( instance, peer )
        if not self._checkModified( value, file_peer ):
            return
        self._copyPeer( file_peer, instance, value )
        setattr( peer, self.name, file_peer )
        
    # implementation details
    @property
    def name( self ):
        return self.context.__name__.lower().replace(' ', '_')
        
    def new( self ):
        return schema.File()

    def _checkModified( self, value, peer ):
        checksum = md5( str(value.data) ).hexdigest()
        if peer.checksum == checksum:
            return False
        return True
        
    def _getPeer( self, instance, peer):
        file_peer = getattr( peer, self.name, None )
        if not file_peer:
            return self.new()
        return file_peer
        
    def _copyPeer( self, file_peer, instance, value ):
        file_peer.attribute = self.name
        file_peer.content = str( value.data )
        file_peer.mime_type = self.context.getContentType( instance ) 
        file_peer.size = len( file_peer.content )
        file_peer.file_name = getattr(value, 'filename', value.getId())
        file_peer.checksum = md5( file_peer.content ).hexdigest()
        

class ImageTransform( FileTransform ):     
    component.adapts( interfaces.IImageField, interfaces.ISchemaTransformer )
    
class PhotoTransform( FileTransform ):     
    component.adapts( interfaces.IPhotoField, interfaces.ISchemaTransformer )    

class BooleanTransform( BaseFieldTransformer):    
    component.adapts( interfaces.IBooleanField, interfaces.ISchemaTransformer )
    column_type = rdb.Boolean

class IntegerTransform( BaseFieldTransformer):
    component.adapts( interfaces.IIntegerField, interfaces.ISchemaTransformer )
    column_type = rdb.Integer
    
class FloatTransform( BaseFieldTransformer ):    
    component.adapts( interfaces.IFloatField, interfaces.ISchemaTransformer )
    column_type = rdb.Float
    
class DateTimeTransform( BaseFieldTransformer ):
    column_type = rdb.DateTime
    component.adapts( interfaces.IDateTimeField, interfaces.ISchemaTransformer )

    def copy( self, instance, peer ):
        value = self.context.getAccessor( instance ) ()
        if not value:
            return
        value = datetime.fromtimestamp( value.timeTime() )
        setattr( peer, self.name, value )
        
class ReferenceTransform( object ):
    
    component.adapts( interfaces.IReferenceField, interfaces.ISchemaTransformer )
    interface.implements( interfaces.IReferenceField )
    
    def __init__( self, context, transformer):
        self.context = context
        self.transformer = transformer
        
    def transform( self ):
        return
        
    def copy( self, instance, peer ):
        value = self.context.getAccessor( instance )()

        single_value = not self.context.multiValued 
        
        if not value:
            return
        
        if not isinstance( value, (list, tuple)):
            value= [ value ]

        rel_map = dict( [ ( (r.relationship, r.target.uid ), r) for r in peer.relations ] )
        oids_seen = set() # oids of the current reference values of this field
        
        for ob in value:
            t_oid = ob.UID()
            oids_seen.add( t_oid )
            
            # skip if the object is already related
            related =  ( self.context.relationship, t_oid ) in rel_map
            if related:
                continue
            # if single value and not the same value, delete previous value
            elif single_value and peer.relations: 
                Session().delete( peer.relations[0] )
            
            # fetch the remote side's peer
            peer_ob = schema.fromUID( ob.UID() )
            if peer_ob is None:
                serializer = interfaces.ISerializer( ob, None )
                if serializer is None: continue
                peer_ob = serializer.add()
                
            # create the relation
            relation = schema.Relation( peer,
                                        peer_ob,
                                        self.context.relationship )

        if single_value:
            return
        
        # delete old values for multi valued relations
        rel_oids = set( [ oid for rel_type, oid in rel_map if rel_type == self.context.relationship ] )
        for oid in (rel_oids - oids_seen):
            Session().delete( rel_map[ ( self.context.relationship, oid)] )
            



# many thanks to mike bayer, for helping me track down this bug fix
# http://www.sqlalchemy.org/trac/ticket/1357

import sqlalchemy
from sqlalchemy.sql import expression

def compare(self, other):
    """Compare this ``_BindParamClause`` to the given clause.
    
    Since ``compare()`` is meant to compare statement syntax, this
    method returns True if the two ``_BindParamClauses`` have just
    the same type.
    """
    return isinstance(other, expression._BindParamClause) and other.type.__class__ == self.type.__class__ \
           and self.value == other.value

if sqlalchemy.__version__ < '0.5.4':
    expression._BindParamClause.compare = compare




import os
from setuptools import setup, find_packages

def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()

setup(
    name="ore.contentmirror",
    version=read('version.txt').strip(),
    url="http://contentmirror.googlecode.com",
    install_requires=['setuptools',
                      'SQLAlchemy>=0.5',
                      'zope.sqlalchemy',
                      'zope.component',
                      'zope.event',
                      'zope.schema',
                      ],
    packages=find_packages(exclude=["*.tests"]),
    namespace_packages=['ore'],
    package_data = {
      '': ['*.txt', '*.zcml', '*.pt'],
    },
    zip_safe=False,
    author='Kapil Thangavelu',
    author_email='kapil.foss@gmail.com',
    description="Deploy/Rescue Content from Plone to a Relational Database",
    long_description=read("ore","contentmirror","readme.txt"),
    license='GPL',
    keywords="plone zope zope3",
    entry_points={
     'console_scripts': ['mirror-ddl = ore.contentmirror.ddl:main',
                         'mirror-bulk = ore.contentmirror.bulk:main'],
        }
    )



####################################################################
# Copyright (c) Kapil Thangavelu. All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
####################################################################

from zope import interface, schema

from zope import component
from zope.app.component.metaconfigure import utility, PublicPermission

import sqlalchemy
from ore.contentmirror import interfaces
from ore.contentmirror import loader

from zope.configuration import fields
from zope import interface

try:
    from zope.component import zcml
except ImportError:
    # Plone 2.5 compatibility
    from zope.app.component import metaconfigure as zcml

class IContentMirror( interface.Interface ):

    content = fields.GlobalObject(
        title=u"Content Class to be Mirrored",
        )

    serializer = fields.GlobalObject(
        title=u"Serializer For the Content",
        required=False
        )

    transformer = fields.GlobalObject(
        title=u"Schema Transformer",
        required=False
        )
    
def mirror( _context, content, serializer=None, transformer=None ):

    if not interfaces.IMirrored.implementedBy( content ):
        interface.classImplements( content, interfaces.IMirrored )

    if serializer:
        zcml.adapter( _context,
                      serializer,
                      interfaces.ISerializer,
                      content,
                      )

    if transformer:
        zcml.adapter( _context,
                      serializer,
                      interfaces.ISchemaTransformer,
                      content,
                      )        

    _context.action(
        discriminator = ( content, interfaces.IMirrored ),
        callable = loader.load,
        args = ( content, )
        )

class IEngineDirective( interface.Interface ):
    """ Creates A Database Engine. Database Engines are named utilities.
    """
    url = schema.URI( title = u'Database URL',
                      description = u'SQLAlchemy Database URL',
                      required = True,
                      )
    
    name = schema.Text( title = u'Engine Name',
                        description = u'Empty if this engine is the default engine.',
                        required = False,
                        default = u'',
                        )
    
    echo = schema.Bool( title = u'Echo SQL statements',
                        description = u'Debugging Echo Log for Engine',
                        required = False,
                        default=False
                        )

    pool_recycle = schema.Int( title = u"Connection Recycle",
                               description=u"Time Given in Seconds",
                               required = False,
                               default = -1
                               )

# keyword arguments to pass to the engine
IEngineDirective.setTaggedValue('keyword_arguments', True)

def engine(_context, url, name='', echo=False, pool_recycle=-1, **kwargs):

    component = sqlalchemy.create_engine( url, echo=echo, pool_recycle=pool_recycle, **kwargs )

    utility( _context,
             provides = interfaces.IDatabaseEngine,
             component = component,
             permission = PublicPermission,
             name = name )
             

class IBindDirective( interface.Interface ):
    """ Binds a MetaData to a database engine.
    """

    engine = schema.Text( title = u"Engine Name" )
    
    metadata = fields.GlobalObject( title=u"Metadata Instance",
                             description = u"Metadata Instance to be bound" )
    
    
def bind( _context, engine, metadata ):

    def _bind( engine_name, metadata ):
        metadata.bind = component.getUtility( interfaces.IDatabaseEngine, engine )

    _context.action(
        discriminator = ('alchemist.bind', metadata ),
        callable = _bind,
        args = ( engine, metadata )
        )




####################################################################
# Copyright (c) Kapil Thangavelu. All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
######################################################################
"""
./bin/zopectl run mirror-batch PORTAL_PATH

a content mirror bulk importer

"""
import transaction, time, sys
from ore.contentmirror import interfaces
from ore.contentmirror.session import Session

def expunge( ob ):
    try: ob._p_deactivate()
    except:pass
        
def main( app ):
    if not len(sys.argv) == 2:
        print "mirror-batch portal_path"
        sys.exit(1)
        
    instance_path = sys.argv[1]
    portal = app.unrestrictedTraverse( instance_path )
     
    threshold = 500
    count = 0
    
    start_time = time.time()
    batch_time = start_time
    
    for brain in portal.portal_catalog(sort_on='created'):

        try: 
            ob = brain.getObject()
            # skip broken objects
            if not ob.UID(): 
                continue
        except: continue
        
        serializer = interfaces.ISerializer( ob, None )
        if serializer is None:
            expunge( ob )
            continue
        
        count += 1

        # the object may have been processed by as a dependency
        # so we use update, which dispatches to add if no db
        # state is found.
        serializer.update()
        expunge( ob )
        
        if count % threshold == 0:

            transaction.commit()
            print "Processed ", count, "in", "%0.2f s"%(time.time()-batch_time)            
            ob._p_jar._cache.incrgc()
            batch_time = time.time()

    # commit the last batch
    transaction.commit()
    print "Processed ", count, "in", "%0.2f s"%(time.time()-batch_time)
    print "Finished in", time.time()-start_time
        
if __name__ == '__main__':
    
    try:
        main( app = app )        
    except:
        import sys,traceback, pdb
        traceback.print_exc()
        pdb.post_mortem(sys.exc_info()[-1])
        raise



####################################################################
# Copyright (c) Kapil Thangavelu. All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
######################################################################


from zope import interface, schema

class IDatabaseEngine( interface.Interface ):
    """
    configuration and access to pooled database connection
    """
                
class IMirrored( interface.Interface ):
    """ marker interface, signifying that the content should be mirrored to a database """

# Base Generic Interface to Attach Default Adapters to    
try:
    from Product.Archetypes.interfaces import IBaseContent as IPortalContent
except:
    class IPortalContent( interface.Interface ): pass

class IMetaData( interface.Interface ):
    """ marker interface for sqlalchemy metadata, to allow for use in adaptation """


# Fields already covered by the base peer content class, common to all content
DUBLIN_CORE= [
    "id",
    "contributors",
    "creators",
    "creation_date",    
    "description",
    "effectiveDate",
    "expirationDate",    
    "language",    
    "location",
    "modification_date",
    "rights",    
    "subject",
    "title",
    'path',
    # the following aren't dublin core but are common to atct, we include
    # them here to have them filtered by default from the peer table
    "allowDiscussion",
    "excludeFromNav",
    ]

#########################################
# Runtime Serialization
#########################################
      
class IOperation( interface.Interface ):

    oid = schema.ASCIILine( description=u"The identifier for the content" )
    filtered = schema.Bool( description=u"set by a filter to denote the opertion should be skipped")
    
    def process( connection ):
        """
        process an index operation
        """
        
class IFilter( interface.Interface ):
    """ a subscription adapter that can filter objects serialized"""

class IDeleteOperation( IOperation ): pass
class IUpdateOperation( IOperation ): pass
class IAddOperation( IOperation ): pass       

class IOperationBuffer( interface.Interface ):
    """ transactional operation buffer """

    def add( operation ):
        """ add an operation to the buffer """

    def flush( ):
        """ flush the buffer and process/handle the operations """

class IOperationBufferFactory( interface.Interface ):
    """ a factory for creating operation buffer instances, as buffers
        are accessed via utilities """

class IOperationFactory( interface.Interface ):
        """ """

class ISerializer( interface.Interface ):

    def add( ):
        """
        add the object to the database
        """
        
    def delete( ):
        """
        delete the object from the database
        """
        
    def update( ):
        """
        update the object state in the database
        """

########################################    
## Schema Generation
########################################
class IModelLoader( interface.Interface ):
    
    def load( klass ):
        """ load a content class"""
        
    def transform( ):
        """ return a schema transformer """
        
    def peer( transformer ):
        """ generate a content peer class for the content class """
        
        
class ISchemaTransformer( interface.Interface ):
    """ translate an archetypes schema to a relational schema """
    
    metadata = interface.Attribute("metadata")
    
    def transform( ):
        """
        return a sqlalchemy database table representation
        """
        
class IFieldTransformer(interface.Interface):
    """ transforms an archetypes field into a sqlalchemy field """
    
    def transform( ):
        """
        """
        
    def copy( instance, peer ):
        """
        """
########################################    
## Content Peers
########################################
#         
class IContentPeer( interface.Interface ):
    """ a rdb persisted class that has a mirror of attributes of a portal content class"""

    transformer = schema.Object( ISchemaTransformer )
    
class IContentFile( interface.Interface ):
    pass

class IPeerFactory( interface.Interface ):
    """ """
    def make( ):
        """
        create a peer class, with a mapper, returns the mapped orm class
        """
        
class IPeerRegistry( interface.Interface ):
    """ a registry mapping a content class to its orm peer class """
    
########################################    
## Interface Specifications for AT Fields
######################################## 

class IComputedField( interface.Interface): pass
class IStringField( interface.Interface ): pass
class IBooleanField( interface.Interface ): pass   
class IIntegerField( interface.Interface ): pass
class IFloatField( interface.Interface ): pass
class IReferenceField( interface.Interface): pass
class ILinesField( interface.Interface ): pass
class IFileField( interface.Interface ): pass
class IImageField( interface.Interface ): pass
class IPhotoField( interface.Interface ): pass
class ITextField( interface.Interface ): pass
class IDateTimeField( interface.Interface ): pass
class IDisplayList( interface.Interface ): pass
